import argparse
import importlib

_module_names = [
    "fasta",
    "taxonomy"
]

def main():
    modules = {}

    parser = argparse.ArgumentParser("dnadb")
    subparsers = parser.add_subparsers(title="module", required=True, dest="module")
    for module_name in _module_names:
        modules[module_name] = importlib.import_module(f".{module_name}", "dnadb.cli")
        modules[module_name].define_arguments(subparsers.add_parser(module_name))
    config = parser.parse_args()
    getattr(modules[config.module], f"command_{config.command.replace('-', '_')}")(config)



import bisect
from dataclasses import dataclass, field, replace
import enum
from functools import singledispatchmethod
import io
import json
import numpy as np
import numpy.typing as npt
from pathlib import Path
import re
from tqdm import tqdm
from typing import Dict, Generator, Iterable, Iterator, List, Literal, Optional, overload, Tuple, TypeVar, Union

from .db import DbFactory, DbWrapper
from .fasta import FastaDb, FastaEntry
from .utils import open_file, sort_dict

RANKS = ("Domain", "Phylum", "Class", "Order", "Family", "Genus", "Species")
RANK_PREFIXES = ''.join(rank[0] for rank in RANKS).lower()

# Utility Functions --------------------------------------------------------------------------------

def is_taxonomy(taxonomy: str) -> bool:
    """
    Check if a string is a valid taxonomy label.
    """
    return bool(re.match(r"^\w__[^;]+(;\s*\w__[^;]*)*;?$", taxonomy))


def split_taxonomy(taxonomy: str, keep_empty: bool = False) -> Tuple[str, ...]:
    """
    Split taxonomy label into a tuple
    """
    return tuple(re.findall(r"\w__([^;]*)" if keep_empty else r"\w__([^;]+)", taxonomy))


def join_taxonomy(taxonomy: Union[Tuple[str, ...], List[str]], depth: Optional[int] = None) -> str:
    """
    Merge a taxonomy tuple into a string format
    """
    if depth is None:
        depth = len(taxonomy)
    assert depth >= 1 and depth <= len(RANKS), "Invalid taxonomy"
    taxonomy = tuple(taxonomy) + ("",)*(depth - len(taxonomy))
    return "; ".join([f"{RANK_PREFIXES[i]}__{taxon}" for i, taxon in enumerate(taxonomy)])


@overload
def taxonomy_parent(taxonomy: str) -> str: ...
@overload
def taxonomy_parent(taxonomy: Tuple[str, ...]) -> Tuple[str, ...]: ...
def taxonomy_parent(taxonomy: Union[str, Tuple[str, ...]]) -> Union[str, Tuple[str, ...]]:
    """
    Get the parent taxonomy of a taxonomy label
    """
    if isinstance(taxonomy, str):
        taxons = split_taxonomy(taxonomy, keep_empty=True)
    else:
        taxons = taxonomy
    # find the last non-empty taxon
    for i in range(len(taxons) - 1, -1, -1):
        if len(taxons[i]) > 0:
            taxons = taxons[:i] + ("",)*(len(taxons) - i)
            break
    if isinstance(taxonomy, str):
        return join_taxonomy(taxons)
    return taxons

# Taxonomy DB --------------------------------------------------------------------------------------

class ITaxonomyEntry:
    sequence_id: str
    label: str

@dataclass(frozen=True, order=True)
class TaxonomyEntry(ITaxonomyEntry):
    sequence_id: str
    label: str

    @property
    def taxons(self):
        return split_taxonomy(self.label, keep_empty=True)

    @property
    def depth(self):
        return len(self.taxons)

    def trim(self, depth: int):
        return replace(self, label=join_taxonomy(self.taxons, depth=depth))

    def __len__(self) -> int:
        return len(self.label)

    def __str__(self) -> str:
        return "\t".join([self.sequence_id, self.label])


TaxonomyDict = Dict[str, "TaxonomyDict"]
class TaxonomyTreeFactory:
    def __init__(self, depth: int = 7):
        self.depth = depth
        self._tree: TaxonomyDict = {}

    def add_taxons(self, taxons: Tuple[str, ...]):
        taxons = (taxons + ('',)*(self.depth - len(taxons)))[:self.depth]
        tree = self._tree
        for taxon in taxons:
            if taxon not in tree:
                tree[taxon] = {}
            tree = tree[taxon]

    def add_label(self, label: str):
        self.add_taxons(split_taxonomy(label, keep_empty=True))

    def add_entry(self, entry: TaxonomyEntry):
        self.add_label(entry.label)

    def add_entries(self, entries: Iterable[TaxonomyEntry]):
        for entry in entries:
            self.add_entry(entry)

    def _sort_tree_dict(self, tree: TaxonomyDict):
        sort_dict(tree)
        for value in tree.values():
            self._sort_tree_dict(value)

    def build(self) -> "TaxonomyTree":
        self._sort_tree_dict(self._tree)
        return TaxonomyTree(self.depth, self._tree)


class TaxonomyTree:

    @dataclass(frozen=True)
    class Taxon:
        taxon_label: str = field(compare=False, hash=False)
        rank: int = field(compare=True, hash=False)
        taxon_id: int = field(compare=True)
        taxonomy_id: int = field(compare=False, hash=False)
        parent: "TaxonomyTree.Taxon" = field(compare=False, hash=False)
        children: Dict[int, "TaxonomyTree.Taxon"] = field(default_factory=dict, compare=False, hash=False)
        child_ids: Dict[str, int] = field(default_factory=dict, compare=False, hash=False)

        def __init__(self, taxon_label: str, taxon_id: int = -1, taxonomy_id: int = -1, parent: Optional["TaxonomyTree.Taxon"] = None):
            object.__setattr__(self, "taxon_label", taxon_label)
            object.__setattr__(self, "rank", parent.rank+1 if parent is not None else -1)
            object.__setattr__(self, "taxon_id", taxon_id)
            object.__setattr__(self, "taxonomy_id", taxonomy_id)
            object.__setattr__(self, "parent", parent) # None is allowed
            object.__setattr__(self, "children", {})
            object.__setattr__(self, "child_ids", {})
            if self.parent is not None:
                self.parent.children[taxon_id] = self
                self.parent.child_ids[taxon_label] = self.taxon_id

        def add_child(self, taxon_label: str, taxon_id: int, taxonomy_id: int) -> "TaxonomyTree.Taxon":
            assert taxon_label not in self.child_ids, f"Taxon {repr(taxon_label)} already exists"
            self.children[taxon_id] = TaxonomyTree.Taxon(taxon_label, taxon_id, taxonomy_id, self)
            self.child_ids[taxon_label] = taxon_id
            return self.children[taxon_id]

        @property
        def taxonomy_label(self) -> str:
            return join_taxonomy(self.taxons)

        @property
        def taxons(self) -> Tuple[str, ...]:
            head = self
            taxons: Tuple[str, ...] = ()
            while head.rank != -1:
                taxons = (head.taxon_label,) + taxons
                head = head.parent
            return taxons

        @property
        def taxon_ids(self) -> Tuple[int, ...]:
            head = self
            taxon_ids: Tuple[int, ...] = ()
            while head.rank != -1:
                taxon_ids = (head.taxon_id,) + taxon_ids
                head = head.parent
            return taxon_ids

        @property
        def taxonomy_ids(self) -> Tuple[int, ...]:
            head = self
            taxonomy_ids: Tuple[int, ...] = ()
            while head.rank != -1:
                taxonomy_ids = (head.taxonomy_id,) + taxonomy_ids
                head = head.parent
            return taxonomy_ids

        def __contains__(self, label_or_taxon_id: Union[str, int]) -> bool:
            if isinstance(label_or_taxon_id, str):
                return label_or_taxon_id in self.child_ids
            return label_or_taxon_id in self.children

        def __getitem__(self, label_or_taxon_id: Union[str, int]) -> "TaxonomyTree.Taxon":
            if isinstance(label_or_taxon_id, str):
                label_or_taxon_id = self.child_ids[label_or_taxon_id]
            return self.children[label_or_taxon_id]

        def __iter__(self) -> Iterator["TaxonomyTree.Taxon"]:
            return iter(self.children.values())

        def __len__(self) -> int:
            return len(self.children)

        def __repr__(self) -> str:
            params = [
                "taxon_label=" + repr(self.taxon_label),
            ]
            if self.rank != -1:
                params += [
                    "taxon_id=" + repr(self.taxon_id),
                    "taxonomy_id=" + repr(self.taxonomy_id),
                    "taxonomy_label=" + repr(self.taxonomy_label),
                ]
            return f"Taxon({', '.join(params)})"

    @classmethod
    def deserialize(cls, taxonomy_tree_bytes: bytes) -> "TaxonomyTree":
        return cls(**json.loads(taxonomy_tree_bytes))

    def __init__(self, depth: int, tree: TaxonomyDict):
        self.depth = depth
        self.id_to_taxon_map, self.taxon_to_id_map = self._build_taxon_id_maps(tree)
        self.tree, self.taxonomy_id_map = self._build_tree_and_taxonomy_map(tree)

    def _build_taxon_id_maps(
        self,
        tree: TaxonomyDict
    ) -> Tuple[Tuple[List[str], ...], Tuple[Dict[str, int], ...]]:
        taxon_id_to_taxon_map: Tuple[List[str], ...] = tuple([] for _ in range(self.depth))
        stack: List[Tuple[int, TaxonomyDict]] = [(0, tree)]
        while len(stack) > 0:
            rank, tree = stack.pop()
            for taxon in tree:
                index = bisect.bisect_left(taxon_id_to_taxon_map[rank], taxon)
                if index >= len(taxon_id_to_taxon_map[rank]) or taxon_id_to_taxon_map[rank][index] != taxon:
                    taxon_id_to_taxon_map[rank].insert(index, taxon)
                if len(tree[taxon]) > 0:
                    stack.append((rank + 1, tree[taxon]))
        taxon_to_taxon_id_map = tuple({t: i for i, t in enumerate(g)} for g in taxon_id_to_taxon_map)
        return taxon_id_to_taxon_map, taxon_to_taxon_id_map

    def _build_tree_and_taxonomy_map(self, tree: TaxonomyDict) -> Tuple[Taxon, Tuple[List[Taxon], ...]]:
        taxonomy_id_to_taxon_map = tuple([] for _ in range(self.depth))
        root = TaxonomyTree.Taxon("Root")
        stack: List[Tuple[TaxonomyTree.Taxon, TaxonomyDict]] = [(root, tree)]
        while len(stack) > 0:
            parent, head = stack.pop()
            s = []
            for label in head:
                taxon_id = self.taxon_to_id_map[parent.rank + 1][label]
                taxonomy_id = len(taxonomy_id_to_taxon_map[parent.rank+1])
                taxon = parent.add_child(label, taxon_id, taxonomy_id)
                taxonomy_id_to_taxon_map[parent.rank+1].append(taxon)
                assert taxonomy_id_to_taxon_map[parent.rank+1][taxonomy_id] == taxon
                if len(head[taxon.taxon_label]) > 0:
                    s.append((taxon, head[taxon.taxon_label]))
            stack += reversed(s)
        return root, taxonomy_id_to_taxon_map

    def reduce_entry(self, label: TaxonomyEntry) -> TaxonomyEntry:
        return replace(label, label=self.reduce_label(label.label))

    def reduce_label(self, label: str) -> str:
        return join_taxonomy(self.reduce_taxons(split_taxonomy(label, keep_empty=True)))

    def reduce_taxons(self, taxons: Tuple[str, ...], pad: bool = True) -> Tuple[str, ...]:
        count = 0
        head = self.tree
        for taxon in taxons:
            if taxon not in head.child_ids:
                break
            count += 1
            head = head[taxon]
        if pad:
            return taxons[:count] + ("",)*(len(taxons) - count)
        return taxons[:count]

    def reduce_taxonomy(
        self,
        taxonomy: Union[TaxonomyEntry, str, int, Tuple[str, ...], Tuple[int, ...]]
    ) -> Union["TaxonomyTree.Taxon", None]:
        if isinstance(taxonomy, int):
            return self.taxonomy_id_map[-1][taxonomy]
        if isinstance(taxonomy, TaxonomyEntry):
            taxonomy = taxonomy.label
        if isinstance(taxonomy, str):
            taxonomy = split_taxonomy(taxonomy, keep_empty=True)
        head = self.tree
        for taxon in taxonomy:
            if taxon not in head:
                break
            head = head[taxon]
        if head.rank == -1:
            return None
        return head

    def has_taxonomy(
        self,
        taxonomy: Union[TaxonomyEntry, str, int, Tuple[str, ...], Tuple[int, ...]]
    ) -> bool:
        if isinstance(taxonomy, int):
            return taxonomy < len(self.taxonomy_id_map[self.depth - 1])
        if isinstance(taxonomy, TaxonomyEntry):
            taxonomy = taxonomy.label
        if isinstance(taxonomy, str):
            taxonomy = split_taxonomy(taxonomy, keep_empty=True)
        head = self.tree
        try:
            for taxon in taxonomy:
                head = head[taxon]
            if head.rank == -1 > 0:
                raise KeyError()
        except KeyError:
            return False
        return True

    def taxonomy(
        self,
        taxonomy: Union[TaxonomyEntry, str, int, Tuple[str, ...], Tuple[int, ...]]
    ) -> "TaxonomyTree.Taxon":
        if isinstance(taxonomy, int):
            return self.taxonomy_id_map[-1][taxonomy]
        if isinstance(taxonomy, TaxonomyEntry):
            taxonomy = taxonomy.label
        if isinstance(taxonomy, str):
            taxonomy = split_taxonomy(taxonomy, keep_empty=True)
        head = self.tree
        try:
            for taxon in taxonomy:
                head = head[taxon]
            if head.rank == -1 > 0:
                raise KeyError()
        except KeyError:
            raise KeyError(f"Taxonomy {taxonomy} not found")
        return head

    def serialize(self) -> bytes:
        tree: TaxonomyDict = {}
        stack: List[Tuple[TaxonomyTree.Taxon, TaxonomyDict]] = [(self.tree, tree)]
        while len(stack) > 0:
            head, tree_head = stack.pop()
            for child in head:
                tree_head[child.taxon_label] = {}
                stack.append((child, tree_head[child.taxon_label]))
        return json.dumps(dict(depth=self.depth, tree=tree)).encode()

    def __contains__(
        self,
        taxonomy: Union[TaxonomyEntry, str, int, Tuple[str, ...], Tuple[int, ...]]
    ) -> bool:
        return self.has_taxonomy(taxonomy)

    def __getitem__(
        self,
        taxonomy: Union[TaxonomyEntry, str, int, Tuple[str, ...], Tuple[int, ...]]
    ) -> "TaxonomyTree.Taxon":
        return self.taxonomy(taxonomy)

    def __len__(self) -> int:
        return len(self.taxonomy_id_map[-1])

    def __iter__(self) -> Iterator["TaxonomyTree.Taxon"]:
        return iter(self.taxonomy_id_map[-1])

    def __eq__(self, other: "TaxonomyTree"):
        return self.serialize() == other.serialize()

    def sample(self, shape: Union[int, Tuple[int, ...]], rng: np.random.Generator) -> np.ndarray:
        result = np.empty(np.product(shape), dtype=object)
        result[:] = list(map(
            lambda i: self.taxonomy_id_map[-1][i],
            rng.choice(len(self), len(result), replace=True)))
        return result.reshape(shape)


@dataclass(frozen=True)
class TaxonomyDbEntry(ITaxonomyEntry):
    db: "TaxonomyDb"
    sequence_index: int
    label_id: int

    @property
    def fasta_entry(self) -> FastaEntry:
        assert self.db.fasta_db is not None, "FASTA DB is not available."
        return self.db.fasta_db[self.sequence_index]

    @property
    def label(self):
        return self.taxonomy.taxonomy_label

    @property
    def sequence_id(self) -> str:
        return self.db.sequence_index_to_id(self.sequence_index)

    @property
    def taxonomy(self) -> "TaxonomyTree.Taxon":
        return self.db.tree.taxonomy_id_map[-1][self.label_id]

    def __repr__(self) -> str:
        return f"TaxonomyDbEntry(sequence_id={repr(self.sequence_id)}, label={repr(self.label)})"


class TaxonomyDbFactory(DbFactory):
    def __init__(self, path: Union[str, Path], fasta_db: FastaDb, depth: int = 7):
        super().__init__(path)
        self.depth = depth
        self.fasta_db = fasta_db
        self.sequences: Dict[str, List[int]] = {}
        self.num_sequences = 0

    def write_sequence(self, sequence_id: str, label: str):
        sequence_index = self.fasta_db.sequence_id_to_index(sequence_id)
        if label not in self.sequences:
            self.sequences[label] = []
        bisect.insort(self.sequences[label], sequence_index)
        self.num_sequences += 1

    def write_entry(self, entry: TaxonomyEntry):
        self.write_sequence(entry.sequence_id, entry.label)

    def write_entries(self, entries: Iterable[TaxonomyEntry]):
        for entry in entries:
            self.write_entry(entry)

    def before_close(self):
        tree = TaxonomyTreeFactory(self.depth)
        for label in self.sequences:
            tree.add_label(label)
        tree = tree.build()
        self.write("fasta_uuid", self.fasta_db.uuid.bytes)
        self.write("tree", tree.serialize())
        self.write("num_sequences", np.int32(self.num_sequences).tobytes())
        self.write("num_labels", np.int32(len(self.sequences)).tobytes())
        for label, sequence_indices in tqdm(self.sequences.items(), desc="Writing labels disk..."):
            taxonomy_id = tree.taxonomy(label).taxonomy_id
            self.write(f"sequences_{taxonomy_id}", np.array(sequence_indices, dtype=np.int32).tobytes())
            for sequence_index in sequence_indices:
                sequence_id = self.fasta_db.index_to_sequence_id(sequence_index)
                self.write(str(sequence_index), np.int32(taxonomy_id).tobytes())
                self.write(f"sequence_index_{sequence_index}", sequence_id.encode())
                self.write(f"sequence_{sequence_id}", np.int32(sequence_index).tobytes())
        return super().before_close()


class TaxonomyDb(DbWrapper):

    # sequence_index -> label_id
    # sequence_index_x -> sequence_id
    # sequence_x -> sequence_index

    class InMemory(enum.Flag):
        Nothing = 0
        SequencesWithTaxonomy = enum.auto()
        SequenceLabels = enum.auto()
        SequenceIdMaps = enum.auto()
        All = SequencesWithTaxonomy | SequenceLabels | SequenceIdMaps

    _sequences_with_label: List[npt.NDArray[np.int32]]|None = None # label_id -> sequence_indices
    _sequence_labels: Dict[int, int]|None = None                   # sequence_index -> label_id
    _sequence_id_to_index: Dict[str, int]|None = None              # sequence_id -> sequence_index
    _sequence_index_to_id: Dict[int, str]|None = None              # sequence_index -> sequence_id

    def __init__(
        self,
        path: Union[str, Path],
        fasta_db: Optional[FastaDb] = None,
        in_memory: InMemory = InMemory.Nothing
    ):
        super().__init__(path)
        self.fasta_db = fasta_db
        if self.fasta_db is not None:
            assert self.fasta_db.uuid.bytes == self.db["fasta_uuid"], "FASTA DB UUID does not match"
        self.tree = TaxonomyTree.deserialize(self.db["tree"])
        self.num_sequences = np.frombuffer(self.db["num_sequences"], dtype=np.int32)[0]
        self.num_labels = np.frombuffer(self.db["num_labels"], dtype=np.int32)[0]

        if TaxonomyDb.InMemory.SequencesWithTaxonomy in in_memory:
            self._sequences_with_label = []
            for i in range(self.num_labels):
                self._sequences_with_label.append(np.frombuffer(self.db[f"sequences_{i}"], dtype=np.int32))

        if TaxonomyDb.InMemory.SequenceLabels in in_memory:
            self._sequence_labels = {}
            for i in range(self.num_sequences):
                self._sequence_labels[i] = np.frombuffer(self.db[str(i)], dtype=np.int32)[0]

        if TaxonomyDb.InMemory.SequenceIdMaps in in_memory:
            self._sequence_id_to_index = {}
            self._sequence_index_to_id = {}
            for i in range(self.num_sequences):
                sequence_id = self.db[f"sequence_index_{i}"].decode()
                self._sequence_id_to_index[sequence_id] = i
                self._sequence_index_to_id[i] = sequence_id

    def count(
        self,
        taxonomy: Union[TaxonomyEntry, str, int, Tuple[str, ...], Tuple[int, ...]]
    ) -> int:
        return sum(1 for _ in self.sequences_with_taxonomy(taxonomy))

    def has_taxonomy(
        self,
        taxonomy: Union[TaxonomyEntry, str, int, Tuple[str, ...], Tuple[int, ...]]
    ) -> bool:
        return self.tree.has_taxonomy(taxonomy)

    def labels(self, depth: int = -1) -> Iterator[TaxonomyTree.Taxon]:
        return iter(self.tree.taxonomy_id_map[depth])

    def sequence_index_to_id(self, sequence_index: int) -> str:
        if self._sequence_index_to_id is not None:
            return self._sequence_index_to_id[sequence_index]
        return self.db[f"sequence_index_{sequence_index}"].decode()

    def sequence_id_to_index(self, sequence_id: str) -> int:
        if self._sequence_id_to_index is not None:
            return self._sequence_id_to_index[sequence_id]
        return np.frombuffer(self.db[f"sequence_{sequence_id}"], dtype=np.int32)[0]

    @singledispatchmethod
    def __contains__(self, sequence_index: int):
        return sequence_index < self.num_sequences

    @__contains__.register
    def _(self, sequence_id: str):
        if self._sequence_id_to_index is not None:
            return sequence_id in self._sequence_id_to_index
        return f"sequence_{sequence_id}" in self.db

    def sequence_indices_with_taxonomy_id(self, taxonomy_id: int) -> npt.NDArray[np.int32]:
        if self._sequences_with_label is not None:
            return self._sequences_with_label[taxonomy_id]
        return np.frombuffer(self.db[f"sequences_{taxonomy_id}"], dtype=np.int32)

    @singledispatchmethod
    def sequences_with_taxonomy(self, taxonomy_id: int) -> Generator[TaxonomyDbEntry, None, None]:
        for i in self.sequence_indices_with_taxonomy_id(taxonomy_id):
            yield TaxonomyDbEntry(self, i, taxonomy_id)

    @sequences_with_taxonomy.register
    def _(self, taxonomy: TaxonomyTree.Taxon) -> Generator[TaxonomyDbEntry, None, None]:
        return self.sequences_with_taxonomy(taxonomy.taxonomy_id)

    @sequences_with_taxonomy.register
    def _(self, taxonomy_label: str) -> Generator[TaxonomyDbEntry, None, None]:
        return self.sequences_with_taxonomy(self.tree.taxonomy(taxonomy_label))

    @singledispatchmethod
    def __getitem__(self, sequence_index: int) -> TaxonomyDbEntry:
        if self._sequence_labels is not None:
            taxonomy_id = self._sequence_labels[sequence_index]
        else:
            taxonomy_id = np.frombuffer(self.db[f"{sequence_index}"], dtype=np.int32)[0]
        return TaxonomyDbEntry(self, sequence_index, taxonomy_id)

    @__getitem__.register
    def _(self, sequence_id: str) -> TaxonomyDbEntry:
        return self[self.sequence_id_to_index(sequence_id)]

    def __len__(self) -> int:
        return self.num_sequences

    def __iter__(self) -> Iterator[TaxonomyDbEntry]:
        for i in range(self.num_sequences):
            yield self[i]

    def sample(self, shape: Union[int, Tuple[int, ...]], rng: np.random.Generator) -> np.ndarray:
        """
        Sample entries weighted by the number of sequences in each taxonomy.
        """
        result = np.empty(np.product(shape), dtype=object)
        result[:] = list(map(
            lambda i: TaxonomyDbEntry(self, rng.choice(self.sequence_indices_with_taxonomy_id(i)), i), # type: ignore
            rng.choice(self.num_labels, len(result), replace=True)))
        return result.reshape(shape)

T = TypeVar("T", bound=ITaxonomyEntry)
def entries(
    taxonomy: Union[io.TextIOBase, Iterable[T], str, Path],
    header: bool|Literal["auto"] = "auto"
) -> Generator[TaxonomyEntry, None, None]:
    """
    Create an Iterable over a taxonomy file or iterable of taxonomy entries.
    """
    if isinstance(taxonomy, (str, Path)):
        with open_file(taxonomy, 'r') as buffer:
            yield from read(buffer, header=header)
    elif isinstance(taxonomy, io.TextIOBase):
        yield from read(taxonomy, header=header)
    else:
        yield from map(lambda entry: TaxonomyEntry(entry.sequence_id, entry.label), taxonomy)


def read(
    buffer: io.TextIOBase,
    header: bool|Literal["auto"] = "auto"
) -> Generator[TaxonomyEntry, None, None]:
    """
    Read taxonomies from a tab-separated file (TSV)
    """
    iterator = iter(buffer)
    if header == "auto":
        sequence_id, taxonomy = next(iterator).strip().split('\t')
        if is_taxonomy(taxonomy):
            yield TaxonomyEntry(sequence_id, taxonomy)
    elif header:
        next(iterator)
    for line in iterator:
        sequence_id, taxonomy = line.rstrip().split('\t')
        yield TaxonomyEntry(sequence_id, taxonomy)


def write(buffer: io.TextIOBase, entries: Iterable[ITaxonomyEntry]):
    """
    Write taxonomy entries to a tab-separate file (TSV)
    """
    for entry in entries:
        buffer.write(f"{entry.sequence_id}\t{entry.label}\n")




from dataclasses import dataclass
import io
import numpy as np
import numpy.typing as npt
from pathlib import Path
from typing import Generator, Iterable, Tuple, Union

from .db import DbFactory, DbWrapper
from .dna import AbstractSequenceWrapper
from .sample import ISample
from .types import int_t
from .utils import open_file

def phred_encode(probabilities: npt.ArrayLike, encoding: int_t = 33) -> str:
    scores = (-10 * np.log10(np.array(probabilities))).astype(int)
    return ''.join((chr(score + encoding)) for score in scores)


def phred_decode(qualities: str, encoding: int_t = 33) -> npt.NDArray[np.float64]:
    scores = np.array([(ord(token) - encoding) for token in qualities])
    return 10**(scores / -10)


@dataclass(frozen=True, order=True)
class FastqHeader:
    """
    A class representation of the header of a FASTQ entry.
    """
    __slots__ = (
        "instrument",
        "run_number",
        "flowcell_id",
        "lane",
        "tile",
        "pos",
        "read_type",
        "is_filtered",
        "control_number",
        "sequence_index"
    )

    instrument: str
    run_number: int_t
    flowcell_id: str
    lane: int_t
    tile: int_t
    pos: Tuple[int, int]
    read_type: int_t
    is_filtered: bool
    control_number: int_t
    sequence_index: str

    @classmethod
    def deserialize(cls, sequence_id: bytes):
        return cls.from_str(sequence_id.decode())

    @classmethod
    def from_str(cls, sequence_id: str) -> "FastqHeader":
        # Split up the sequence ID information
        left, right = sequence_id.strip()[1:].split(' ')
        left = left.split(':')
        right = right.split(':')
        return cls(
            instrument=left[0],
            run_number=int(left[1]),
            flowcell_id=left[2],
            lane=int(left[3]),
            tile=int(left[4]),
            pos=(int(left[5]), int(left[6])),
            read_type=int(right[0]),
            is_filtered=right[1] == 'Y',
            control_number=int(right[2]),
            sequence_index=right[3]
        )

    # Serialize a FastqHeader object to a byte string
    def serialize(self) -> bytes:
        return str(self).encode()

    def __str__(self):
        sequence_id = '@'
        sequence_id += ':'.join(map(str, [
            self.instrument,
            self.run_number,
            self.flowcell_id,
            self.lane,
            self.tile,
            *self.pos
        ]))
        sequence_id += ' '
        sequence_id += ':'.join(map(str, [
            self.read_type,
            'Y' if self.is_filtered else 'N',
            self.control_number,
            self.sequence_index
        ]))
        return sequence_id


@dataclass(frozen=True, order=True)
class FastqEntry(AbstractSequenceWrapper):
    """
    A class representation of a FASTQ entry containing the sequnce identifier, sequence, and quality
    scores.
    """
    __slots__ = ("header_str", "quality_scores")

    header_str: str
    quality_scores: str

    @classmethod
    def deserialize(cls, entry: bytes) -> "FastqEntry":
        return cls(*entry.decode().split('\x00'))

    @classmethod
    def from_str(cls, entry: str) -> "FastqEntry":
        header, sequence, _, quality_scores= entry.rstrip().split('\n')
        return cls(header, sequence, quality_scores)

    def __init__(self, header: str, sequence: str, quality_scores: str):
        object.__setattr__(self, "header_str", header)
        object.__setattr__(self, "sequence", sequence)
        object.__setattr__(self, "quality_scores", quality_scores)

    def serialize(self) -> bytes:
        return '\x00'.join((self.header_str, self.sequence, self.quality_scores)).encode()

    @property
    def header(self):
        return FastqHeader.from_str(self.header_str)

    def __str__(self):
        return f"{self.header_str}\n{self.sequence}\n+\n{self.quality_scores}"


class FastqDbFactory(DbFactory):
    """
    A factory for creating LMDB-backed databases of FASTA entries.
    """
    def __init__(self, path: Union[str, Path], chunk_size: int_t = 10000):
        super().__init__(path, chunk_size)
        self.num_entries = np.int32(0)

    def write_entry(self, entry: FastqEntry):
        """
        Create a new FASTA LMDB database from a FASTA file.
        """
        self.write(str(self.num_entries), entry.serialize())
        self.num_entries += 1

    def write_entries(self, entries: Iterable[FastqEntry]):
        for entry in entries:
            self.write_entry(entry)

    def before_close(self):
        self.write("length", self.num_entries.tobytes())
        super().before_close()


class FastqDb(ISample[FastqEntry], DbWrapper):
    def __init__(self, fastq_db_path: Union[str, Path]):
        super().__init__(fastq_db_path)
        self.length = np.frombuffer(self.db["length"], dtype=np.int32, count=1)[0]

    def __len__(self):
        return self.length

    def __contains__(self, sequence_index: int_t) -> bool:
        return sequence_index < self.length

    def __iter__(self) -> Generator[FastqEntry, None, None]:
        for i in range(len(self)):
            yield self[i]

    def __getitem__(self, sequence_index: int_t) -> FastqEntry:
        return FastqEntry.deserialize(self.db[str(sequence_index)])

    def sample(self, shape: Union[int, Tuple[int, ...]], rng: np.random.Generator) -> np.ndarray:
        """
        Sample sequences from the FASTA database.
        """
        result = np.empty(np.product(shape), dtype=object)
        result[:] = list(map(self.__getitem__, rng.choice(self.length, len(result), replace=True)))
        return result.reshape(shape)


def entries(
    sequences: Union[io.TextIOBase, Iterable[FastqEntry], str, Path]
) -> Iterable[FastqEntry]:
    """
    Create an iterator over a FASTQ file or iterable of FASTQ entries.
    """
    if isinstance(sequences, (str, Path)):
        with open_file(sequences, 'r') as buffer:
            yield from read(buffer)
    elif isinstance(sequences, io.TextIOBase):
        yield from read(sequences)
    else:
        yield from sequences


def read(buffer: io.TextIOBase) -> Generator[FastqEntry, None, None]:
    """
    Read entries from a FASTQ file buffer.
    """
    line = buffer.readline()
    while len(line) > 0:
        entry_str = "".join((line, *(buffer.readline() for _ in range(3))))
        yield FastqEntry.from_str(entry_str)
        line = buffer.readline()


def write(buffer: io.TextIOBase, entries: Iterable[FastqEntry]) -> int:
    """
    Write entries to a FASTQ file.
    """
    bytes_written = 0
    for entry in entries:
        bytes_written += buffer.write(str(entry) + '\n')
    return bytes_written



import bisect
from dataclasses import dataclass
from functools import singledispatchmethod
import io
import numpy as np
import numpy.typing as npt
from pathlib import Path
from typing import Callable, Generator, Iterable, Tuple, Union

from .db import DbFactory, DbWrapper
from .dna import AbstractSequenceWrapper
from .sample import ISample
from .utils import open_file

@dataclass(frozen=True, order=True)
class FastaEntry(AbstractSequenceWrapper):
    """
    A container class to represent a FASTA entry
    """
    __slots__ = ("identifier", "extra")

    identifier: str
    extra: str

    @classmethod
    def deserialize(cls, entry: bytes) -> "FastaEntry":
        """
        Deserialize a FASTA entry from a byte string
        """
        return cls(*entry.decode().split('\x00'))

    @classmethod
    def from_str(cls, entry: str) -> "FastaEntry":
        """
        Create a FASTA entry from a string
        """
        header, *sequence_parts = entry.split('\n')
        header_line = header[1:].rstrip().split(maxsplit=1)
        identifier = header_line[0]
        extra = header_line[1] if len(header_line) > 1 else ""
        sequence = "".join(sequence_parts)
        return cls(identifier, sequence, extra)

    def __init__(self, identifier: str, sequence: str, extra: str = ""):
        object.__setattr__(self, "identifier", identifier)
        object.__setattr__(self, "sequence", sequence)
        object.__setattr__(self, "extra", extra)

    def serialize(self) -> bytes:
        return "\x00".join((self.identifier, self.sequence, self.extra)).encode()

    def __str__(self):
        header_line = f"{self.identifier} {self.extra}".rstrip()
        return f">{header_line}\n{self.sequence}"


class FastaDbFactory(DbFactory):
    """
    A factory for creating LMDB-backed databases of FASTA entries.
    """
    __slots__ = ("num_entries", "has_ambiguous_bases")

    def __init__(self, path: Union[str, Path], chunk_size: int = 10000):
        super().__init__(path, chunk_size)
        self.num_entries = np.int32(0)
        self.has_ambiguous_bases = False

    def write_entry(self, entry: FastaEntry):
        """
        Create a new FASTA LMDB database from a FASTA file.
        """
        self.write(f"id_{entry.identifier}", np.int32(self.num_entries).tobytes())
        self.write(str(self.num_entries), entry.serialize())
        self.num_entries += 1

    def write_entries(self, entries: Iterable[FastaEntry]):
        for entry in entries:
            self.write_entry(entry)

    def before_close(self):
        self.write("length", self.num_entries.tobytes())
        super().before_close()


class FastaDb(ISample[FastaEntry], DbWrapper):
    """
    An LMDB-backed database of FASTA entries.
    """
    __slots__ = ("length", "contains_sequence_id", "sequence_id_to_index", "entry", "_id_map", "_sequences")

    length: int
    contains_sequence_id: Callable[[str], bool]
    sequence_id_to_index: Callable[[str], int]
    entry: Callable[[int], FastaEntry]

    _id_map: dict[str, int]|None
    _sequences: npt.NDArray[np.object_]|None

    def __init__(
        self,
        fasta_db_path: Union[str, Path],
        load_id_map_into_memory: bool = False,
        load_sequences_into_memory: bool = False,
    ):
        super().__init__(fasta_db_path)
        self.length = np.frombuffer(self.db["length"], dtype=np.int32, count=1)[0]

        if load_sequences_into_memory:
            self._sequences = sequences = np.empty((self.length,), dtype=np.object_)
            for i in range(self.length):
                self._sequences[i] = FastaEntry.deserialize(self.db[str(i)])
            self.entry = lambda index: sequences[index]
        else:
            self._sequences = None
            self.entry = lambda index: FastaEntry.deserialize(self.db[str(index)])

        if load_id_map_into_memory:
            self._id_map = id_map = {self.entry(i).identifier: i for i in range(self.length)}
            self.contains_sequence_id = lambda sequence_id: (sequence_id in id_map) # type: ignore
            self.sequence_id_to_index = lambda sequence_id: id_map[sequence_id] # type: ignore
        else:
            self._id_map = None
            self.contains_sequence_id = lambda sequence_id: f"id_{sequence_id}" in self.db
            self.sequence_id_to_index = lambda sequence_id: np.frombuffer(self.db[f"id_{sequence_id}"], dtype=np.int32, count=1)[0]

    def __len__(self):
        return self.length

    def contains_index(self, sequence_index: int) -> bool:
        return sequence_index < self.length

    def index_to_sequence_id(self, sequence_index: int) -> str:
        return self.entry(sequence_index).identifier

    @singledispatchmethod
    def __contains__(self, sequence_index: int) -> bool:
        return self.contains_index(sequence_index)

    @__contains__.register
    def _(self, sequence_id: str) -> bool:
        return self.contains_sequence_id(sequence_id)

    @__contains__.register
    def _(self, entry: FastaEntry) -> bool:
        return self.contains_sequence_id(entry.identifier)

    def __iter__(self):
        for i in range(len(self)):
            yield self[i]

    @singledispatchmethod
    def __getitem__(self, sequence_index: int) -> FastaEntry:
        return self.entry(sequence_index)

    @__getitem__.register
    def _(self, sequence_id: str) -> FastaEntry:
        return self.entry(self.sequence_id_to_index(sequence_id))

    def mappings(
        self,
        fasta_mapping_db_path: Union[str, Path],
        load_into_memory: bool = False
    ) -> "Tuple[FastaMappingEntry, ...]":
        return FastaMappingDb(fasta_mapping_db_path, self, load_into_memory).entries

    def sample(self, shape: Union[int, Tuple[int, ...]], rng: np.random.Generator) -> np.ndarray:
        """
        Sample sequences from the FASTA database.
        """
        result = np.empty(np.product(shape), dtype=object)
        result[:] = list(map(self.entry, rng.choice(self.length, len(result), replace=True)))
        return result.reshape(shape)


class FastaMappingEntryFactory:
    def __init__(self, name: str, fasta_mapping_db_factory: "FastaMappingDbFactory"):
        self.name = name
        self.fasta_mapping_db_factory = fasta_mapping_db_factory
        self.sequence_indices: list[int] = []

    def write_entry(self, entry: FastaEntry, abundance: int = 1):
        index = self.fasta_mapping_db_factory.fasta_db.sequence_id_to_index(entry.identifier)
        insert_index = bisect.bisect_right(self.sequence_indices, index)
        for i in range(abundance):
            self.sequence_indices.insert(insert_index+i, index)

    def write_entries(self, entries: Iterable[FastaEntry]):
        for entry in entries:
            self.write_entry(entry)

    def __enter__(self) -> "FastaMappingEntryFactory":
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.fasta_mapping_db_factory.write_entry(self)


class FastaMappingEntry(ISample[FastaEntry]):

    _sequence_indices: npt.NDArray[np.uint32]|None

    def __init__(
        self,
        index: int,
        fasta_mapping_db: "FastaMappingDb",
        load_into_memory: bool = False,
    ):
        self.index = index
        self.fasta_mapping_db = fasta_mapping_db
        self.name = self.fasta_mapping_db.db[f"{index}_name"].decode()
        self.length = np.frombuffer(self.fasta_mapping_db.db[f"{index}_length"], dtype=np.int32, count=1)[0]

        if load_into_memory:
            self._sequence_indices = sequence_indices = np.empty(self.length, dtype=np.uint32)
            for i in range(self.length):
                self._sequence_indices[i] = np.frombuffer(self.fasta_mapping_db.db[f"{index}_{i}"], dtype=np.uint32, count=1)[0]
            self.sequence_index = lambda mapped_sequence_index: sequence_indices[mapped_sequence_index]
        else:
            self._sequence_indices = None
            self.sequence_index = lambda mapped_sequence_index: np.frombuffer(self.fasta_mapping_db.db[f"{index}_{mapped_sequence_index}"], dtype=np.uint32, count=1)[0]

    def entry(self, mapped_sequence_index: int) -> FastaEntry:
        return self.fasta_mapping_db.fasta_db.entry(self.sequence_index(mapped_sequence_index))

    @singledispatchmethod
    def __contains__(self, sequence_index: int) -> bool:
        if self._sequence_indices is not None:
            return sequence_index == np.searchsorted(self._sequence_indices, sequence_index)
        low = 0
        high = self.length - 1
        while low <= high:
            mid = (low + high) // 2
            mid_val = self.sequence_index(mid)
            if mid_val < sequence_index:
                low = mid + 1
            elif mid_val > sequence_index:
                high = mid - 1
            else:
                return True
        return False

    @__contains__.register
    def _(self, sequence_id: str) -> bool:
        sequence_index = self.fasta_mapping_db.fasta_db.sequence_id_to_index(sequence_id)
        return sequence_index in self

    @__contains__.register
    def _(self, entry: FastaEntry) -> bool:
        return entry.identifier in self

    @singledispatchmethod
    def __getitem__(self, mapped_sequence_index: int) -> FastaEntry:
        return self.entry(mapped_sequence_index)

    @__getitem__.register
    def _(self, sequence_id: str) -> FastaEntry:
        return self.fasta_mapping_db.fasta_db[sequence_id]

    def __iter__(self) -> Generator[FastaEntry, None, None]:
        for i in range(self.length):
            yield self[i]

    def __len__(self) -> int:
        return self.length

    def sample(self, shape: Union[int, Tuple[int, ...]], rng: np.random.Generator) -> np.ndarray:
        """
        Sample sequences from the FASTA database.
        """
        result = np.empty(np.product(shape), dtype=object)
        result[:] = list(map(self.entry, rng.choice(self.length, len(result), replace=True)))
        return result.reshape(shape)


class FastaMappingDbFactory(DbFactory):
    def __init__(self, path: Union[str, Path], fasta_db: FastaDb, chunk_size: int = 10000):
        super().__init__(path, chunk_size)
        self.fasta_db = fasta_db
        self.num_entries = np.int32(0)
        self.write("fasta_db_uuid", self.fasta_db.uuid.bytes)

    def create_entry(self, name: str) -> FastaMappingEntryFactory:
        return FastaMappingEntryFactory(name, self)

    def write_entry(self, entry: FastaMappingEntryFactory):
        self.write(f"{self.num_entries}_name", entry.name.encode())
        self.write(f"{self.num_entries}_length", np.int32(len(entry.sequence_indices)).tobytes())
        for i, index in enumerate(np.array(entry.sequence_indices, dtype=np.uint32)):
            self.write(f"{self.num_entries}_{i}", index.tobytes())
        self.num_entries += 1

    def write_entries(self, entries: Iterable[FastaMappingEntryFactory]):
        for entry in entries:
            self.write_entry(entry)

    def before_close(self):
        self.write("length", self.num_entries.tobytes())
        super().before_close()


class FastaMappingDb(DbWrapper):
    def __init__(
        self,
        path: Union[str, Path],
        fasta_db: FastaDb,
        load_into_memory: bool = False
    ):
        super().__init__(path)
        assert fasta_db.uuid.bytes == self.db["fasta_db_uuid"], "This FASTA Mapping was not created with the given FASTA DB."
        self.fasta_db = fasta_db
        self.length = np.frombuffer(self.db["length"], dtype=np.int32, count=1)[0]
        self.entries = tuple(FastaMappingEntry(i, self, load_into_memory) for i in range(self.length))

    def __getitem__(self, index: int) -> FastaMappingEntry:
        return self.entries[index]

    def __len__(self):
        return self.length

def entries(
    sequences: Union[io.TextIOBase, Iterable[FastaEntry], str, Path]
) -> Iterable[FastaEntry]:
    """
    Create an iterator over a FASTA file or iterable of FASTA entries.
    """
    if isinstance(sequences, (str, Path)):
        with open_file(sequences, 'r') as buffer:
            yield from read(buffer)
    elif isinstance(sequences, io.TextIOBase):
        yield from read(sequences)
    else:
        yield from sequences


# def entries_with_taxonomy(
#     sequences: Iterable[FastaEntry],
#     taxonomies: Iterable[TaxonomyEntry],
# ) -> Generator[Tuple[FastaEntry, TaxonomyEntry], None, None]:
#     """
#     Efficiently iterate over a FASTA file with a corresponding taxonomy file
#     """
#     labels = {}
#     taxonomy_iterator = iter(taxonomies)
#     taxonomy: TaxonomyEntry
#     for sequence in sequences:
#         while sequence.identifier not in labels:
#             taxonomy = next(taxonomy_iterator)
#             labels[taxonomy.identifier] = taxonomy
#         taxonomy = labels[sequence.identifier]
#         del labels[sequence.identifier]
#         yield sequence, taxonomy


def read(buffer: io.TextIOBase) -> Generator[FastaEntry, None, None]:
    """
    Read entries from a FASTA file buffer.
    """
    entry_str = buffer.readline()
    for line in buffer:
        if line.startswith('>'):
            yield FastaEntry.from_str(entry_str)
            entry_str = ""
        entry_str += line
    if len(entry_str) > 0:
        yield FastaEntry.from_str(entry_str)


def write(buffer: io.TextIOBase, entries: Iterable[FastaEntry]) -> int:
    """
    Write entries to a FASTA file.
    """
    bytes_written = 0
    for entry in entries:
        bytes_written += buffer.write(str(entry) + '\n')
    return bytes_written



import argparse
from dataclasses import replace
from pathlib import Path
from tqdm import tqdm
from typing import Callable, Dict

def define_arguments(parser: argparse.ArgumentParser):
    subparsers = parser.add_subparsers(required=True, dest="command")

    import_args = subparsers.add_parser(
        "import",
        help="Convert a taxonomy TSV file to a taxonomy TSV DB.")
    import_args.add_argument(
        "--depth",
        type=int,
        default=None)
    import_args.add_argument(
        "--fasta-db",
        type=Path,
        required=True)
    import_args.add_argument(
        "input_path",
        type=Path,
        help="The taxonomy TSV file to convert.")
    import_args.add_argument(
        "output_path",
        type=Path,
        nargs='?',
        default=None,
        help="The destination taxonomy TSV DB file.")

    export_args = subparsers.add_parser(
        "export",
        help="Convert a taxonomy TSV DB to a taxonomy TSV file.")
    export_args.add_argument(
        "input_path",
        type=Path,
        help="The taxonomy TSV DB file to convert.")
    export_args.add_argument(
        "output_path",
        type=Path,
        nargs='?',
        default=None,
        help="The destination taxonomy TSV file.")

    info_args = subparsers.add_parser(
        "info",
        help="Display information about a taxonomy TSV or TSV DB file.")
    info_args.add_argument(
        "input_path",
        type=Path,
        help="The taxonomy TSV or TSV DB file to display information about.")

    lookup_args = subparsers.add_parser(
        "lookup",
        help="Lookup A FASTA ID in the given taxonomy TSV or TSV DB.")
    lookup_args.add_argument(
        "input_path",
        type=Path,
        help="The taxonomy TSV or TSV DB file to lookup the FASTA ID in.")
    lookup_args.add_argument(
        "ids",
        nargs='+',
        type=str,
        help="The FASTA ID(s) to lookup.")


def command_import(config: argparse.Namespace):
    print("Importing taxonomy TSV...")
    from dnadb import fasta, taxonomy
    output_path = config.output_path
    if output_path is None:
        output_path = config.input_path.with_suffix(".tsv.db")
    fasta_db = fasta.FastaDb(config.fasta_db)
    num_skipped = 0
    num_processed = 0
    with taxonomy.TaxonomyDbFactory(output_path, fasta.FastaDb(config.fasta_db)) as factory:
        for entry in tqdm(taxonomy.entries(config.input_path)):
            if entry.sequence_id not in fasta_db:
                num_skipped += 1
                continue
            if config.depth is not None:
                taxons = taxonomy.split_taxonomy(entry.label, keep_empty=True)[:config.depth]
                entry = replace(entry, label=taxonomy.join_taxonomy(taxons, depth=config.depth))
            factory.write_entry(entry)
            num_processed += 1
    print(f"Done. Imported {num_processed:,} sequences. Skipped {num_skipped:,} sequences.")


def command_export(config: argparse.Namespace):
    print("Exporting taxonomy TSV DB...")
    from dnadb import taxonomy
    output_path = config.output_path
    if output_path is None:
        output_path = config.input_path.with_suffix("") # Remove .db suffix
    with open(output_path, "w") as output, taxonomy.TaxonomyDb(config.input_path) as db:
        output.write("Feature ID\tTaxon\n")
        taxonomy.write(output, tqdm(db))


def command_info(config: argparse.Namespace):
    from dnadb import taxonomy
    uuid = None
    if config.input_path.suffix == ".db":
        db = taxonomy.TaxonomyDb(config.input_path)
        uuid = db.uuid
        count = sum(db.count(label_index) for label_index in range(len(db)))
        unique_labels = len(db)
    else:
        entries = taxonomy.entries(config.input_path)
        count = 0
        unique_labels = set()
        for entry in entries:
            count += 1
            unique_labels.add(entry.label)
        unique_labels = len(unique_labels)

    print(f"Info for: {config.input_path}")
    if uuid is not None:
        print(f"               UUID: {uuid}")
    print(f"             Length: {count:,}")
    print(f"  Unique Taxonomies: {unique_labels:,}")


def command_lookup(config: argparse.Namespace):
    from dnadb import taxonomy
    if config.input_path.suffix == ".db":
        db = taxonomy.TaxonomyDb(config.input_path)
        for id in config.ids:
            if id not in db:
                print(f"'>{id}' not found.")
            else:
                print(f"{id}\t{db[id]}")
    else:
        entries = taxonomy.entries(config.input_path)
        entries_to_print: Dict[str, str|None] = {id: None for id in config.ids}
        found = 0
        for entry in entries:
            if entry.sequence_id in entries_to_print:
                found += 1
                entries_to_print[entry.sequence_id] = entry.label
                if found == len(entries_to_print):
                    break
        for id, entry in entries_to_print.items():
            if entry is None:
                print(f"'{id}' not found.")
            else:
                print(entry)



import numpy as np
import numpy.typing as npt
import tensorflow as tf
from typing import Optional

from ... import dna

BASE_LOOKUP_TABLE = tf.constant(dna.BASE_LOOKUP_TABLE, dtype=tf.int32)
BASE_REVERSE_LOOKUP_TABLE = tf.constant(dna.BASE_REVERSE_LOOKUP_TABLE, dtype=tf.int32)
IUPAC_AUGMENT_LOOKUP_TABLE = tf.constant(dna.IUPAC_AUGMENT_LOOKUP_TABLE, dtype=tf.int32)

# DNA Sequence Encoding/Decoding -------------------------------------------------------------------

def encode(
    sequences: str|bytes|npt.NDArray[np.str_]|tf.Tensor,
    kmer: Optional[int|tf.Tensor] = 1,
    ambiguous_bases: bool = False,
    augment_ambiguous_bases: bool = True
) -> tf.Tensor:
    """
    Encode the given DNA sequences.
    """
    result = _encode_bases(sequences)
    if ambiguous_bases and augment_ambiguous_bases:
        result = augment_ambiguous_bases(result)
    if kmer > 1:
        result = encode_kmers(result, kmer, ambiguous_bases)
    return result


def decode(
    sequences: str|bytes|npt.NDArray[np.str_]|tf.Tensor,
    kmer: Optional[int|tf.Tensor] = 1,
    ambiguous_bases: bool = False
) -> tf.Tensor:
    """
    Decode the given DNA sequences.
    """
    if kmer > 1:
        sequences = decode_kmers(ambiguous_bases)
    return _decode_bases(sequences)

def _encode_bases(sequences: str|bytes|npt.NDArray[np.str_]|tf.Tensor) -> tf.Tensor:
    """
    Encode a DNA sequence into an integer vector representation.
    """
    ascii = tf.cast(tf.io.decode_raw(sequences, tf.uint8), tf.int32)
    return tf.gather(BASE_LOOKUP_TABLE, ascii - 65)


def _decode_bases(sequences: tf.Tensor) -> tf.Tensor:
    """
    Decode a DNA sequence integer vector representation into a string of bases.
    """
    ascii = tf.gather(BASE_REVERSE_LOOKUP_TABLE, sequences)
    return tf.strings.unicode_encode(ascii, output_encoding="UTF-8")


def _encode_kmers(
    sequences: tf.Tensor,
    kmer: int|tf.Tensor,
    ambiguous_bases: bool = False
) -> tf.Tensor:
    """
    Convert DNA sequences into sequences of k-mers.
    """
    original_shape = tf.shape(sequences)
    sequence_length = tf.shape(sequences)[-1]
    sequences = tf.reshape(sequences, (-1, sequence_length, 1))
    num_bases = len(dna.BASES + (dna.AMBIGUOUS_BASES if ambiguous_bases else ""))
    kernel = tf.reshape(tf.pow(num_bases, tf.range(kmer-1, -1, -1, dtype=tf.int32)), (kmer, 1, 1))
    result = tf.nn.convolution(sequences, kernel, padding="VALID")
    return tf.reshape(
        result,
        tf.concat((original_shape[:-1], (sequence_length - kmer + 1,)), axis=0))


def _decode_kmers(
    kmer_sequences: tf.Tensor,
    kmer: int|tf.Tensor,
    ambiguous_bases: bool = False
) -> tf.Tensor:
    """
    Decode sequence of k-mers into 1-mer DNA sequences.
    """
    num_bases = len(dna.BASES + (dna.AMBIGUOUS_BASES if ambiguous_bases else ""))
    powers = tf.range(kmer - 1, -1, -1)
    kernel = num_bases**powers
    return tf.concat([
        kmer_sequences // kernel[0],
        tf.repeat(kmer_sequences[:,-1:], kmer - 1, axis=-1) % kernel[:-1] // kernel[1:]
    ], axis=-1)


def _augment_ambiguous_bases(sequences: tf.Tensor):
    sequences = tf.expand_dims(sequences, -1)
    return tf.gather_nd(
        IUPAC_AUGMENT_LOOKUP_TABLE,
        tf.concat((
            sequences,
            tf.random.uniform(tf.shape(sequences), minval=0, maxval=len(dna.ALL_BASES), dtype=tf.int32)),
            axis=-1))


from dataclasses import dataclass
import itertools
import numpy as np
import numpy.typing as npt
import scipy as sp

BASES = "ACGT"
AMBIGUOUS_BASES = "MRWSYKVHDBN"
ALL_BASES = BASES + AMBIGUOUS_BASES

# Map each ambiguous base to a tuple of bases as per the IUAPC standard.
# https://droog.gs.washington.edu/parc/images/iupac.html
# https://iubmb.qmul.ac.uk/misc/naseq.html
IUPAC_MAP = {__b: __c for __b, __c in zip(ALL_BASES, [
    __c for __n in range(1, 5) for __c in itertools.combinations(BASES, __n)])}

# Hashtables ---------------------------------------------------------------------------------------

# A lookup table for converting bases to integers.
# The index is the ASCII value of the base minus the ASCII value of 'A'.
BASE_LOOKUP_TABLE = np.full(np.max(list(map(ord, ALL_BASES))) - ord('A') + 1, 255, dtype=np.uint8)
for __i, __base in enumerate(ALL_BASES):
    BASE_LOOKUP_TABLE[ord(__base) - ord('A')] = __i

# A reverse lookup table for converting integers to bases.
BASE_REVERSE_LOOKUP_TABLE = np.array(list(map(ord, ALL_BASES)), dtype=np.uint8)

# A lookup table for augmenting ambiguous bases to concrete bases.
__iupac_lcm = np.lcm.reduce([len(v) for v in IUPAC_MAP.values()])
IUPAC_AUGMENT_LOOKUP_TABLE = np.full((len(IUPAC_MAP), __iupac_lcm), 255, dtype=np.uint8)
for __base, __bases in IUPAC_MAP.items():
    __repeat = __iupac_lcm // len(__bases)
    __repeated_bases = BASE_LOOKUP_TABLE[np.array(list(map(ord, __bases*__repeat)), np.uint8) - ord('A')]
    IUPAC_AUGMENT_LOOKUP_TABLE[BASE_LOOKUP_TABLE[ord(__base) - ord('A')]] = __repeated_bases

# DNA Sequence Encoding/Decoding -------------------------------------------------------------------

def encode(ascii_bases: npt.NDArray[np.uint8]) ->  npt.NDArray[np.uint8]:
    """
    Encode the given DNA bases in ASCII form into an integer vector representation.
    """
    return BASE_LOOKUP_TABLE[ascii_bases - 65]


def decode(sequences: npt.NDArray[np.uint8]) -> npt.NDArray[np.uint8]:
    """
    Decoed the given DNA bases into ASCII an ascii vector representation.
    """
    return BASE_REVERSE_LOOKUP_TABLE[sequences]


def encode_sequence(sequence: str) -> npt.NDArray[np.uint8]:
    """
    Encode a DNA sequence into an integer vector representation.
    """
    return encode(np.frombuffer(sequence.encode(), np.uint8))


def decode_sequence(sequence: npt.NDArray[np.uint8]) -> str:
    """
    Decode a DNA sequence integer vector representation into a string of bases.
    """
    return decode(sequence).tobytes().decode()


def encode_kmers(
    sequences: npt.NDArray[np.uint8],
    kmer: int,
    ambiguous_bases: bool = False
) -> npt.NDArray[np.int64]:
    """
    Convert DNA sequences into sequences of k-mers.
    """
    slices = [slice(0, s) for s in sequences.shape[:-1]]
    edge_slices = slice((kmer - 1) // 2, (kmer - 1) // -2)
    num_bases = len(BASES + (AMBIGUOUS_BASES if ambiguous_bases else ""))
    powers = np.arange(kmer).reshape((1,)*len(slices) + (-1,))
    kernel = num_bases**powers
    return sp.ndimage.convolve(sequences, kernel)[(*slices, edge_slices)]


def decode_kmers(
    sequences: np.ndarray,
    kmer: int,
    ambiguous_bases: bool = False
) -> npt.NDArray[np.uint8]:
    """
    Decode sequence of k-mers into 1-mer DNA sequences.
    """
    slices = [slice(0, s) for s in sequences.shape[:-1]]
    edge_slice = slice(-1, sequences.shape[-1])
    num_bases = len(BASES + (AMBIGUOUS_BASES if ambiguous_bases else ""))
    powers = np.arange(kmer - 1, -1, -1)
    kernel = num_bases**powers
    edge = (sequences[(*slices, edge_slice)] % kernel[:-1]) // kernel[1:]
    return np.concatenate([sequences // kernel[0], edge], axis=-1).astype(np.uint8)


def augment_ambiguous_bases(
    sequence: str,
    rng: np.random.Generator = np.random.default_rng()
) -> str:
    """
    Replace the ambiguous bases in a DNA sequence at random with a valid concrete base.
    """
    return decode_sequence(replace_ambiguous_encoded_bases(encode_sequence(sequence), rng))


def replace_ambiguous_encoded_bases(
    encoded_sequences: npt.NDArray[np.uint8],
    rng: np.random.Generator = np.random.default_rng()
) -> npt.NDArray[np.uint8]:
    """
    Replace the ambiguous bases in an encoded DNA sequence at random with a valid concrete base.
    """
    augment_indices = rng.integers(0, 12, size=encoded_sequences.shape)
    return IUPAC_AUGMENT_LOOKUP_TABLE[encoded_sequences, augment_indices]


def to_rna(dna_sequence: str) -> str:
    """
    Convert an RNA sequence to DNA.
    """
    return dna_sequence.replace('T', 'U')


def to_dna(rna_sequence: str) -> str:
    """
    Convert a DNA sequence to RNA.
    """
    return rna_sequence.replace('U', 'T')

# Data Classes -------------------------------------------------------------------------------------

@dataclass(frozen=True)
class AbstractSequenceWrapper:
    __slots__ = ("sequence",)

    sequence: str

    def __len__(self):
        return len(self.sequence)



from lmdbm import Lmdb
from pathlib import Path
from typing import TypeVar, Union
import uuid

from .types import int_t

T = TypeVar("T")

class DbFactory:
    """
    A wrapper around lmdbm.Lmdb that allows for buffered writes.
    """
    def __init__(self, path: Union[str, Path], chunk_size: int_t = 10000):
        self.path = Path(path)
        if self.path.suffix != ".db":
            self.path = Path(str(self.path) + ".db")
        self.db = Lmdb.open(str(self.path), "n", lock=True)
        self.buffer: dict[Union[str, bytes], bytes] = {}
        self.chunk_size = chunk_size
        self.is_closed = False
        self.uuid = uuid.uuid4()
        self.write("uuid", self.uuid.bytes)

    def flush(self):
        self.db.update(self.buffer)
        self.buffer.clear()

    def contains(self, key: Union[str, bytes]) -> bool:
        return key in self.buffer or key in self.db

    def read(self, key: Union[str, bytes]) -> bytes:
        return self.buffer[key] if key in self.buffer else self.db[key]

    def append(self, key: Union[str, bytes], value: bytes):
        self.write(key, self.read(key) + value)

    def write(self, key: Union[str, bytes], value: bytes):
        self.buffer[key] = value
        if len(self.buffer) >= self.chunk_size:
            self.flush()

    def before_close(self):
        self.flush()

    def close(self):
        if self.is_closed:
            return
        self.is_closed = True
        self.before_close()
        self.db.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def __del__(self):
        self.close()


class DbWrapper:
    __slots__ = ("_path", "_db", "_is_closed", "_uuid")

    def __init__(self, path: Union[str, Path]):
        self._path = Path(path).absolute()
        self._db = Lmdb.open(str(path), lock=False)
        self._is_closed = False
        try:
            self._uuid = uuid.UUID(bytes=self.db["uuid"])
        except KeyError:
            raise Exception(f"Database at {self.path} does not have a UUID.")

    def close(self):
        if self._is_closed:
            return
        self._is_closed = True
        return self._db.close()

    @property
    def db(self) -> Lmdb:
        return self._db

    @property
    def path(self) -> Path:
        return self._path

    @property
    def uuid(self) -> uuid.UUID:
        return self._uuid

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def __del__(self):
        self.close()



import gzip
import io
from pathlib import Path
import requests
import subprocess
from tqdm.auto import tqdm
from typing import cast, Union


def download(url: str, destination: Union[str, Path], chunk_size: int = 1024):
    """
    Download a file from the internet to the provided destination.
    """
    # https://gist.github.com/yanqd0/c13ed29e29432e3cf3e7c38467f42f51
    response = requests.get(url, stream=True)
    if response.status_code != 200:
        raise Exception(f"Could not download: {url}\n{response.content}")
    total = int(response.headers.get('content-length', 0))
    with open(str(destination), 'wb') as file, tqdm(
        desc=f"Downloading: {destination}",
        total=total,
        unit='iB',
        unit_scale=True,
        unit_divisor=1024,
        leave=False
    ) as bar:
        for data in response.iter_content(chunk_size=chunk_size):
            size = file.write(data)
            bar.update(size)


def compress(path: Union[str, Path]):
    """
    Compress the given file with gzip.
    """
    subprocess.run(["gzip", "-f", str(path)])


def decompress(path: Union[str, Path]):
    """
    Decompress the given file with gunzip.
    """
    subprocess.run(["gunzip", "-f", str(path)])


def open_file(path: Union[str, Path], mode: str = "r") -> io.TextIOWrapper:
    """
    Open a file without worrying about compression.
    """
    path = Path(path)
    if path.suffix == ".gz":
        if mode == 'r':
            mode = 'rt'
        elif mode == 'w':
            mode = 'wt'
        return cast(io.TextIOWrapper, gzip.open(path, mode))
    return cast(io.TextIOWrapper, open(path, mode))


def sort_dict(d: dict):
    """
    Sort a dictionary in place.
    """
    for key in sorted(d):
        value = d[key]
        del d[key]
        d[key] = value



import abc
import numpy as np
from typing import Generator, Generic, Tuple, TypeVar, Union

from .dna import AbstractSequenceWrapper

T = TypeVar("T", bound=AbstractSequenceWrapper)
class ISample(abc.ABC, Generic[T]):

    name: str

    def sample(self, shape: Union[int, Tuple[int, ...]], rng: np.random.Generator) -> T:
        """
        Sample sequences from the sample.
        """
        raise NotImplementedError()

    def __contains__(self, sequence_index: int) -> bool:
        """
        Check whether a sequence entry is in the sample.
        """
        raise NotImplementedError()

    def __getitem__(self, sequence_index: int) -> T:
        """
        Get a sequence entry by its index.
        """
        raise NotImplementedError()

    def __iter__(self) -> Generator[T, None, None]:
        """
        Iterate over all sequence entries in the sample.
        """
        raise NotImplementedError()

    def __len__(self) -> int:
        """
        Get the number of sequence entries in the sample.
        """
        raise NotImplementedError()



import argparse
from pathlib import Path
from tqdm import tqdm
from typing import Dict

def define_arguments(parser: argparse.ArgumentParser):
    subparsers = parser.add_subparsers(required=True, dest="command")

    import_args = subparsers.add_parser(
        "import",
        help="Convert a FASTA file to a FASTA DB.")
    import_args.add_argument(
        "--min-length",
        type=int,
        default=0,
        help="The minimum length of a sequence to import.")
    import_args.add_argument(
        "input_path",
        type=Path,
        help="The FASTA file to convert.")
    import_args.add_argument(
        "output_path",
        type=Path,
        nargs='?',
        default=None,
        help="The destination FASTA DB file.")

    export_args = subparsers.add_parser(
        "export",
        help="Convert a FASTA DB to a FASTA file.")
    export_args.add_argument(
        "input_path",
        type=Path,
        help="The FASTA DB file to convert.")
    export_args.add_argument(
        "output_path",
        type=Path,
        nargs='?',
        default=None,
        help="The destination FASTA file.")

    info_args = subparsers.add_parser(
        "info",
        help="Display information about a FASTA or FASTA DB file.")
    info_args.add_argument(
        "input_path",
        type=Path,
        help="The FASTA or FASTA DB file to display information about.")

    lookup_args = subparsers.add_parser(
        "lookup",
        help="Lookup A FASTA ID in the given FASTA or FASTA DB.")
    lookup_args.add_argument(
        "input_path",
        type=Path,
        help="The FASTA or FASTA DB file to lookup the FASTA ID in.")
    lookup_args.add_argument(
        "ids",
        nargs='+',
        type=str,
        help="The FASTA ID(s) to lookup.")


def command_import(config: argparse.Namespace):
    print("Importing FASTA...")
    from dnadb import fasta
    output_path = config.output_path
    if output_path is None:
        output_path = config.input_path.with_suffix(".fasta.db")
    num_skipped = 0
    num_processed = 0
    with fasta.FastaDbFactory(output_path) as factory:
        for entry in tqdm(fasta.entries(config.input_path)):
            if len(entry.sequence) < config.min_length:
                num_skipped += 1
                continue
            factory.write_entry(entry)
            num_processed += 1
    print(f"Done. Imported {num_processed:,} sequences. Skipped {num_skipped:,} sequences.")


def command_export(config: argparse.Namespace):
    print("Exporting FASTA DB...")
    from dnadb import fasta
    output_path = config.output_path
    if output_path is None:
        output_path = config.input_path.with_suffix("") # Remove .db suffix
    with open(output_path, "w") as output, fasta.FastaDb(config.input_path) as db:
        fasta.write(output, tqdm(db))


def command_info(config: argparse.Namespace):
    from dnadb import fasta
    uuid = None
    if config.input_path.suffix == ".db":
        fasta_db = fasta.FastaDb(config.input_path)
        uuid = fasta_db.uuid
        entries = iter(fasta_db)
    else:
        entries = fasta.entries(config.input_path)
    min_length = float("inf")
    max_length = float("-inf")
    count = 0
    for entry in entries:
        count += 1
        length = len(entry.sequence)
        min_length = min(min_length, length)
        max_length = max(max_length, length)
    print(f"Info for: {config.input_path}")
    if uuid is not None:
        print(f"               UUID: {uuid}")
    print(f"             Length: {count:,}")
    print(f"  Shortest Sequence: {min_length:,}")
    print(f"   Longest Sequence: {max_length:,}")


def command_lookup(config: argparse.Namespace):
    from dnadb import fasta
    if config.input_path.suffix == ".db":
        db = fasta.FastaDb(config.input_path)
        for id in config.ids:
            if id not in db:
                print(f"'>{id}' not found.")
            else:
                print(db[id])
    else:
        entries = fasta.entries(config.input_path)
        entries_to_print: Dict[str, fasta.FastaEntry|None] = {id: None for id in config.ids}
        found = 0
        for entry in entries:
            if entry.identifier in entries_to_print:
                found += 1
                entries_to_print[entry.identifier] = entry
                if found == len(entries_to_print):
                    break
        for id, entry in entries_to_print.items():
            if entry is None:
                print(f"'{id}' not found.")
            else:
                print(entry)
